@@ -1,3 +1,6 @@
+# pylint:  disable=too-many-lines
+# (Yes, it has a point!)
+
 from __future__ import division, absolute_import, print_function
 
 __copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"
@@ -25,21 +28,15 @@ THE SOFTWARE.
 
 import operator
 import sys
-
-from pytools.decorator import decorator
-import six
-from six.moves import range, zip, intern, input
+import logging
 from functools import reduce
 
-try:
-    decorator_module = __import__("decorator", level=0)
-except TypeError:
-    # this must be Python 2.4
-    my_decorator = decorator
-except ImportError:
-    my_decorator = decorator
-else:
-    my_decorator = decorator_module.decorator
+import six
+from six.moves import range, zip, intern, input
+
+
+decorator_module = __import__("decorator", level=0)
+my_decorator = decorator_module.decorator
 
 __doc__ = """
 A Collection of Utilities
@@ -123,12 +120,30 @@ Name generation
 
 .. autofunction:: generate_unique_names
 .. autofunction:: generate_numbered_unique_names
-.. autofunction:: UniqueNameGenerator
+.. autoclass:: UniqueNameGenerator
 
 Functions for dealing with (large) auxiliary files
 --------------------------------------------------
 
 .. autofunction:: download_from_web_if_not_present
+
+Helpers for :mod:`numpy`
+------------------------
+
+.. autofunction:: reshaped_view
+
+
+Timing data
+-----------
+
+.. autoclass:: ProcessTimer
+
+Log utilities
+-------------
+
+.. autoclass:: ProcessLogger
+.. autoclass:: DebugProcessLogger
+.. autoclass:: log_process
 """
 
 
@@ -217,9 +232,12 @@ class RecordWithoutPickling(object):
 
     __slots__ = []
 
-    def __init__(self, valuedict=None, exclude=["self"], **kwargs):
+    def __init__(self, valuedict=None, exclude=None, **kwargs):
         assert self.__class__ is not Record
 
+        if exclude is None:
+            exclude = ["self"]
+
         try:
             fields = self.__class__.fields
         except AttributeError:
@@ -260,6 +278,13 @@ class RecordWithoutPickling(object):
 
         fields.update(new_fields)
 
+    def __getattr__(self, name):
+        # This method is implemented to avoid pylint 'no-member' errors for
+        # attribute access.
+        raise AttributeError(
+                "'%s' object has no attribute '%s'" % (
+                    self.__class__.__name__, name))
+
 
 class Record(RecordWithoutPickling):
     __slots__ = []
@@ -316,39 +341,6 @@ class Reference(object):
         self.value = value
 
 
-# {{{ dictionary with default
-
-class DictionaryWithDefault(object):
-    def __init__(self, default_value_generator, start={}):
-        self._Dictionary = dict(start)
-        self._DefaultGenerator = default_value_generator
-
-    def __getitem__(self, index):
-        try:
-            return self._Dictionary[index]
-        except KeyError:
-            value = self._DefaultGenerator(index)
-            self._Dictionary[index] = value
-            return value
-
-    def __setitem__(self, index, value):
-        self._Dictionary[index] = value
-
-    def __contains__(self, item):
-        return True
-
-    def iterkeys(self):
-        return six.iterkeys(self._Dictionary)
-
-    def __iter__(self):
-        return self._Dictionary.__iter__()
-
-    def iteritems(self):
-        return six.iteritems(self._Dictionary)
-
-# }}}
-
-
 class FakeList(object):
     def __init__(self, f, length):
         self._Length = length
@@ -368,7 +360,10 @@ class FakeList(object):
 # {{{ dependent dictionary ----------------------------------------------------
 
 class DependentDictionary(object):
-    def __init__(self, f, start={}):
+    def __init__(self, f, start=None):
+        if start is None:
+            start = {}
+
         self._Function = f
         self._Dictionary = start.copy()
 
@@ -377,7 +372,7 @@ class DependentDictionary(object):
 
     def __contains__(self, key):
         try:
-            self[key]
+            self[key]  # pylint: disable=pointless-statement
             return True
         except KeyError:
             return False
@@ -511,15 +506,15 @@ def memoize(*args, **kwargs):
             # http://www.phyast.pitt.edu/~micheles/python/
             key = key_func(*args, **kwargs)
             try:
-                return func._memoize_dic[key]
+                return func._memoize_dic[key]  # pylint: disable=protected-access
             except AttributeError:
                 # _memoize_dic doesn't exist yet.
                 result = func(*args, **kwargs)
-                func._memoize_dic = {key: result}
+                func._memoize_dic = {key: result}  # pylint: disable=protected-access
                 return result
             except KeyError:
                 result = func(*args, **kwargs)
-                func._memoize_dic[key] = result
+                func._memoize_dic[key] = result  # pylint: disable=protected-access
                 return result
     else:
         @my_decorator
@@ -527,15 +522,15 @@ def memoize(*args, **kwargs):
             # by Michele Simionato
             # http://www.phyast.pitt.edu/~micheles/python/
             try:
-                return func._memoize_dic[args]
+                return func._memoize_dic[args]  # pylint: disable=protected-access
             except AttributeError:
                 # _memoize_dic doesn't exist yet.
                 result = func(*args)
-                func._memoize_dic = {args: result}
+                func._memoize_dic = {args: result}  # pylint:disable=protected-access
                 return result
             except KeyError:
                 result = func(*args)
-                func._memoize_dic[args] = result
+                func._memoize_dic[args] = result  # pylint: disable=protected-access
                 return result
     if not args:
         return _deco
@@ -586,10 +581,9 @@ def memoize_on_first_arg(function, cache_dict_name=None):
     def clear_cache(obj):
         delattr(obj, cache_dict_name)
 
-    if sys.version_info >= (2, 5):
-        from functools import update_wrapper
-        new_wrapper = update_wrapper(wrapper, function)
-        new_wrapper.clear_cache = clear_cache
+    from functools import update_wrapper
+    new_wrapper = update_wrapper(wrapper, function)
+    new_wrapper.clear_cache = clear_cache
 
     return new_wrapper
 
@@ -604,13 +598,18 @@ def memoize_method(method):
     return memoize_on_first_arg(method, intern("_memoize_dic_"+method.__name__))
 
 
-def memoize_method_with_uncached(uncached_args=[], uncached_kwargs=set()):
+def memoize_method_with_uncached(uncached_args=None, uncached_kwargs=None):
     """Supports cache deletion via ``method_name.clear_cache(self)``.
 
     :arg uncached_args: a list of argument numbers
         (0-based, not counting 'self' argument)
     """
 
+    if uncached_args is None:
+        uncached_args = []
+    if uncached_kwargs is None:
+        uncached_kwargs = set()
+
     # delete starting from the end
     uncached_args = sorted(uncached_args, reverse=True)
     uncached_kwargs = list(uncached_kwargs)
@@ -764,7 +763,7 @@ def monkeypatch_method(cls):
     return decorator
 
 
-def monkeypatch_class(name, bases, namespace):
+def monkeypatch_class(_name, bases, namespace):
     # from GvR, http://mail.python.org/pipermail/python-dev/2008-January/076194.html
 
     assert len(bases) == 1, "Exactly one base class required"
@@ -810,13 +809,13 @@ def len_iterable(iterable):
     return sum(1 for i in iterable)
 
 
-def flatten(list):
+def flatten(iterable):
     """For an iterable of sub-iterables, generate each member of each
     sub-iterable in turn, i.e. a flattened version of that super-iterable.
 
     Example: Turn [[a,b,c],[d,e,f]] into [a,b,c,d,e,f].
     """
-    for sublist in list:
+    for sublist in iterable:
         for j in sublist:
             yield j
 
@@ -827,7 +826,7 @@ def general_sum(sequence):
 
 def linear_combination(coefficients, vectors):
     result = coefficients[0] * vectors[0]
-    for c, v in zip(coefficients, vectors)[1:]:
+    for c, v in zip(coefficients[1:], vectors[1:]):
         result += c*v
     return result
 
@@ -840,8 +839,8 @@ def common_prefix(iterable, empty=None):
         return empty
 
     for v in it:
-        for j in range(len(pfx)):
-            if pfx[j] != v[j]:
+        for j, pfx_j in enumerate(pfx):
+            if pfx_j != v[j]:
                 pfx = pfx[:j]
                 if j == 0:
                     return pfx
@@ -850,14 +849,14 @@ def common_prefix(iterable, empty=None):
     return pfx
 
 
-def decorate(function, list):
-    return [(x, function(x)) for x in list]
+def decorate(function, iterable):
+    return [(x, function(x)) for x in iterable]
 
 
-def partition(criterion, list):
+def partition(criterion, iterable):
     part_true = []
     part_false = []
-    for i in list:
+    for i in iterable:
         if criterion(i):
             part_true.append(i)
         else:
@@ -881,8 +880,8 @@ def product(iterable):
     return reduce(mul, iterable, 1)
 
 
-all = six.moves.builtins.all
-any = six.moves.builtins.any
+all = six.moves.builtins.all  # pylint: disable=redefined-builtin
+any = six.moves.builtins.any  # pylint: disable=redefined-builtin
 
 
 def reverse_dictionary(the_dict):
@@ -936,7 +935,7 @@ def find_max_where(predicate, prec=1e-5, initial_guess=1, fail_bound=1e38):
 
     # {{{ establish bracket
 
-    mag = 1
+    mag = initial_guess
 
     if predicate(mag):
         mag *= 2
@@ -977,8 +976,8 @@ def find_max_where(predicate, prec=1e-5, initial_guess=1, fail_bound=1e38):
             lower_true = mid
         else:
             upper_false = mid
-    else:
-        return lower_true
+
+    return lower_true
 
     # }}}
 
@@ -1069,16 +1068,16 @@ def average(iterable):
     it = iterable.__iter__()
 
     try:
-        sum = next(it)
+        s = next(it)
         count = 1
     except StopIteration:
         raise ValueError("empty average")
 
     for value in it:
-        sum = sum + value
+        s = s + value
         count += 1
 
-    return sum/count
+    return s/count
 
 
 class VarianceAggregator:
@@ -1095,9 +1094,9 @@ class VarianceAggregator:
 
     def step(self, x):
         self.n += 1
-        delta = x - self.mean
-        self.mean += delta/self.n
-        self.m2 += delta*(x - self.mean)
+        delta_ = x - self.mean
+        self.mean += delta_/self.n
+        self.m2 += delta_*(x - self.mean)
 
     def finalize(self):
         if self.entire_pop:
@@ -1139,7 +1138,7 @@ def indices_in_shape(shape):
     if isinstance(shape, int):
         shape = (shape,)
 
-    if len(shape) == 0:
+    if not shape:
         yield ()
     elif len(shape) == 1:
         for i in range(0, shape[0]):
@@ -1154,7 +1153,7 @@ def indices_in_shape(shape):
 def generate_nonnegative_integer_tuples_below(n, length=None, least=0):
     """n may be a sequence, in which case length must be None."""
     if length is None:
-        if len(n) == 0:
+        if not n:
             yield ()
             return
 
@@ -1177,23 +1176,24 @@ def generate_nonnegative_integer_tuples_below(n, length=None, least=0):
             yield my_part + base
 
 
-def generate_decreasing_nonnegative_tuples_summing_to(n, length, min=0, max=None):
+def generate_decreasing_nonnegative_tuples_summing_to(
+        n, length, min_value=0, max_value=None):
     if length == 0:
         yield ()
     elif length == 1:
-        if n <= max:
-            #print "MX", n, max
+        if n <= max_value:
+            #print "MX", n, max_value
             yield (n,)
         else:
             return
     else:
-        if max is None or n < max:
-            max = n
+        if max_value is None or n < max_value:
+            max_value = n
 
-        for i in range(min, max+1):
+        for i in range(min_value, max_value+1):
             #print "SIG", sig, i
             for remainder in generate_decreasing_nonnegative_tuples_summing_to(
-                    n-i, length-1, min, i):
+                    n-i, length-1, min_value, i):
                 yield (i,) + remainder
 
 
@@ -1260,10 +1260,10 @@ def generate_permutations(original):
     if len(original) <= 1:
         yield original
     else:
-        for perm in generate_permutations(original[1:]):
-            for i in range(len(perm)+1):
+        for perm_ in generate_permutations(original[1:]):
+            for i in range(len(perm_)+1):
                 #nb str[0:1] works in both string and list contexts
-                yield perm[:i] + original[0:1] + perm[i:]
+                yield perm_[:i] + original[0:1] + perm_[i:]
 
 
 def generate_unique_permutations(original):
@@ -1272,10 +1272,10 @@ def generate_unique_permutations(original):
 
     had_those = set()
 
-    for perm in generate_permutations(original):
-        if perm not in had_those:
-            had_those.add(perm)
-            yield perm
+    for perm_ in generate_permutations(original):
+        if perm_ not in had_those:
+            had_those.add(perm_)
+            yield perm_
 
 
 def enumerate_basic_directions(dimensions):
@@ -1295,12 +1295,18 @@ def get_read_from_map_from_permutation(original, permuted):
     Requires that the permutation can be inferred from
     C{original} and C{permuted}.
 
-    >>> for p1 in generate_permutations(range(5)):
-    ...     for p2 in generate_permutations(range(5)):
-    ...         rfm = get_read_from_map_from_permutation(p1, p2)
-    ...         p2a = [p1[rfm[i]] for i in range(len(p1))]
-    ...         assert p2 == p2a
+    .. doctest ::
+
+        >>> for p1 in generate_permutations(range(5)):
+        ...     for p2 in generate_permutations(range(5)):
+        ...         rfm = get_read_from_map_from_permutation(p1, p2)
+        ...         p2a = [p1[rfm[i]] for i in range(len(p1))]
+        ...         assert p2 == p2a
     """
+    from warnings import warn
+    warn("get_read_from_map_from_permutation is deprecated and will be "
+            "removed in 2019", DeprecationWarning, stacklevel=2)
+
     assert len(original) == len(permuted)
     where_in_original = dict(
             (original[i], i) for i in range(len(original)))
@@ -1316,14 +1322,20 @@ def get_write_to_map_from_permutation(original, permuted):
     Requires that the permutation can be inferred from
     C{original} and C{permuted}.
 
-    >>> for p1 in generate_permutations(range(5)):
-    ...     for p2 in generate_permutations(range(5)):
-    ...         wtm = get_write_to_map_from_permutation(p1, p2)
-    ...         p2a = [0] * len(p2)
-    ...         for i, oi in enumerate(p1):
-    ...             p2a[wtm[i]] = oi
-    ...         assert p2 == p2a
+    .. doctest::
+
+        >>> for p1 in generate_permutations(range(5)):
+        ...     for p2 in generate_permutations(range(5)):
+        ...         wtm = get_write_to_map_from_permutation(p1, p2)
+        ...         p2a = [0] * len(p2)
+        ...         for i, oi in enumerate(p1):
+        ...             p2a[wtm[i]] = oi
+        ...         assert p2 == p2a
     """
+    from warnings import warn
+    warn("get_write_to_map_from_permutation is deprecated and will be "
+            "removed in 2019", DeprecationWarning, stacklevel=2)
+
     assert len(original) == len(permuted)
 
     where_in_permuted = dict(
@@ -1337,9 +1349,11 @@ def get_write_to_map_from_permutation(original, permuted):
 
 # {{{ graph algorithms
 
-def a_star(initial_state, goal_state, neighbor_map,
+def a_star(  # pylint: disable=too-many-locals
+        initial_state, goal_state, neighbor_map,
         estimate_remaining_cost=None,
-        get_step_cost=lambda x, y: 1):
+        get_step_cost=lambda x, y: 1
+        ):
     """
     With the default cost and heuristic, this amounts to Dijkstra's algorithm.
     """
@@ -1347,7 +1361,7 @@ def a_star(initial_state, goal_state, neighbor_map,
     from heapq import heappop, heappush
 
     if estimate_remaining_cost is None:
-        def estimate_remaining_cost(x):
+        def estimate_remaining_cost(x):  # pylint: disable=function-redefined
             if x != goal_state:
                 return 1
             else:
@@ -1368,7 +1382,7 @@ def a_star(initial_state, goal_state, neighbor_map,
     queue = [(init_remcost, AStarNode(initial_state, parent=None, path_cost=0))]
     visited_states = set()
 
-    while len(queue):
+    while queue:
         _, top = heappop(queue)
         visited_states.add(top.state)
 
@@ -1430,7 +1444,9 @@ class Table:
                               for col_width in col_widths)]
         return "\n".join(lines)
 
-    def latex(self, skip_lines=0, hline_after=[]):
+    def latex(self, skip_lines=0, hline_after=None):
+        if hline_after is None:
+            hline_after = []
         lines = []
         for row_nr, row in list(enumerate(self.rows))[skip_lines:]:
             lines.append(" & ".join(row)+r" \\")
@@ -1444,7 +1460,8 @@ class Table:
 
 # {{{ histogram formatting
 
-def string_histogram(iterable, min_value=None, max_value=None,
+def string_histogram(  # pylint: disable=too-many-arguments,too-many-locals
+        iterable, min_value=None, max_value=None,
         bin_count=20, width=70, bin_starts=None, use_unicode=True):
     if bin_starts is None:
         if min_value is None or max_value is None:
@@ -1504,7 +1521,7 @@ def word_wrap(text, width, wrap_using="\n"):
     breaks are posix newlines (``\n``).
     """
     space_or_break = [" ", wrap_using]
-    return reduce(lambda line, word, width=width: '%s%s%s' %
+    return reduce(lambda line, word: '%s%s%s' %
             (line,
                 space_or_break[(len(line)-line.rfind('\n')-1
                     + len(word.split('\n', 1)[0])
@@ -1530,7 +1547,11 @@ class CPyUserInterface(object):
     class Parameters(Record):
         pass
 
-    def __init__(self, variables, constants={}, doc={}):
+    def __init__(self, variables, constants=None, doc=None):
+        if constants is None:
+            constants = {}
+        if doc is None:
+            doc = {}
         self.variables = variables
         self.constants = constants
         self.doc = doc
@@ -1558,16 +1579,14 @@ class CPyUserInterface(object):
                 print("    %s" % self.doc[c])
 
     def gather(self, argv=None):
-        import sys
-
         if argv is None:
             argv = sys.argv
 
         if len(argv) == 1 or (
-                ("-h" in argv) or
-                ("help" in argv) or
-                ("-help" in argv) or
-                ("--help" in argv)):
+                ("-h" in argv)
+                or ("help" in argv)
+                or ("-help" in argv)
+                or ("--help" in argv)):
             self.show_usage(argv[0])
             sys.exit(2)
 
@@ -1619,17 +1638,19 @@ class MovedFunctionDeprecationWrapper:
 
 class StderrToStdout(object):
     def __enter__(self):
-        import sys
+        # pylint: disable=attribute-defined-outside-init
         self.stderr_backup = sys.stderr
         sys.stderr = sys.stdout
 
     def __exit__(self, exc_type, exc_val, exc_tb):
-        import sys
         sys.stderr = self.stderr_backup
         del self.stderr_backup
 
 
-def typedump(val, max_seq=5, special_handlers={}):
+def typedump(val, max_seq=5, special_handlers=None):
+    if special_handlers is None:
+        special_handlers = {}
+
     try:
         hdlr = special_handlers[type(val)]
     except KeyError:
@@ -1678,7 +1699,7 @@ def invoke_editor(s, filename="edit.txt", descr="the file"):
     if "EDITOR" in os.environ:
         from subprocess import Popen
         p = Popen([os.environ["EDITOR"], full_name])
-        os.waitpid(p.pid, 0)[1]
+        os.waitpid(p.pid, 0)
     else:
         print("(Set the EDITOR environment variable to be "
                 "dropped directly into an editor next time.)")
@@ -1696,7 +1717,7 @@ def invoke_editor(s, filename="edit.txt", descr="the file"):
 
 # {{{ progress bars
 
-class ProgressBar:
+class ProgressBar(object):  # pylint: disable=too-many-instance-attributes
     """
     .. automethod:: draw
     .. automethod:: progress
@@ -1744,7 +1765,6 @@ class ProgressBar:
             else:
                 eta_str = "?"
 
-            import sys
             sys.stderr.write("%-20s [%s] ETA %s\r" % (
                 self.description,
                 squares*"#"+(self.length-squares)*" ",
@@ -1761,7 +1781,6 @@ class ProgressBar:
         self.draw()
 
     def finished(self):
-        import sys
         self.set_progress(self.total)
         sys.stderr.write("\n")
 
@@ -1783,7 +1802,6 @@ def assert_not_a_file(name):
 
 
 def add_python_path_relative_to_script(rel_path):
-    import sys
     from os.path import dirname, join, abspath
 
     script_name = sys.argv[0]
@@ -1876,7 +1894,10 @@ class UniqueNameGenerator(object):
     .. automethod:: add_names
     .. automethod:: __call__
     """
-    def __init__(self, existing_names=set(), forced_prefix=""):
+    def __init__(self, existing_names=None, forced_prefix=""):
+        if existing_names is None:
+            existing_names = set()
+
         self.existing_names = existing_names.copy()
         self.forced_prefix = forced_prefix
         self.prefix_to_counter = {}
@@ -1918,7 +1939,7 @@ class UniqueNameGenerator(object):
 
         self.prefix_to_counter[based_on] = counter
 
-        var_name = intern(var_name)
+        var_name = intern(var_name)  # pylint: disable=undefined-loop-variable
 
         self.existing_names.add(var_name)
         self._name_added(var_name)
@@ -1934,6 +1955,8 @@ class MinRecursionLimit(object):
         self.min_rec_limit = min_rec_limit
 
     def __enter__(self):
+        # pylint: disable=attribute-defined-outside-init
+
         self.prev_recursion_limit = sys.getrecursionlimit()
         new_limit = max(self.prev_recursion_limit, self.min_rec_limit)
         sys.setrecursionlimit(new_limit)
@@ -1943,15 +1966,13 @@ class MinRecursionLimit(object):
         # (or long chains of to-be gc'd generators) that cannot be
         # undergo garbage collection with a lower recursion limit.
         #
-        # Force a garbage collection while the recursion limit is
-        # still high to avoid this situation.
+        # As a result, it doesn't seem possible to lower the recursion limit
+        # again after it has been raised without causing reliability issues.
         #
         # See https://gitlab.tiker.net/inducer/sumpy/issues/31 for
         # context.
-        import gc
-        gc.collect()
 
-        sys.setrecursionlimit(self.prev_recursion_limit)
+        pass
 
 # }}}
 
@@ -1978,6 +1999,242 @@ def download_from_web_if_not_present(url, local_name=None):
 # }}}
 
 
+# {{{ find git revisions
+
+def find_git_revision(tree_root):  # pylint: disable=too-many-locals
+    # Keep this routine self-contained so that it can be copy-pasted into
+    # setup.py.
+
+    from os.path import join, exists, abspath
+    tree_root = abspath(tree_root)
+
+    if not exists(join(tree_root, ".git")):
+        return None
+
+    # construct minimal environment
+    # stolen from
+    # https://github.com/numpy/numpy/blob/055ce3e90b50b5f9ef8cf1b8641c42e391f10735/setup.py#L70-L92
+    import os
+    env = {}
+    for k in ['SYSTEMROOT', 'PATH', 'HOME']:
+        v = os.environ.get(k)
+        if v is not None:
+            env[k] = v
+    # LANGUAGE is used on win32
+    env['LANGUAGE'] = 'C'
+    env['LANG'] = 'C'
+    env['LC_ALL'] = 'C'
+
+    from subprocess import Popen, PIPE, STDOUT
+    p = Popen(["git", "rev-parse", "HEAD"], shell=False,
+              stdin=PIPE, stdout=PIPE, stderr=STDOUT, close_fds=True,
+              cwd=tree_root, env=env)
+    (git_rev, _) = p.communicate()
+
+    if sys.version_info >= (3,):
+        git_rev = git_rev.decode()
+
+    git_rev = git_rev.rstrip()
+
+    retcode = p.returncode
+    assert retcode is not None
+    if retcode != 0:
+        from warnings import warn
+        warn("unable to find git revision")
+        return None
+
+    return git_rev
+
+
+def find_module_git_revision(module_file, n_levels_up):
+    from os.path import dirname, join
+    tree_root = join(*([dirname(module_file)] + [".." * n_levels_up]))
+
+    return find_git_revision(tree_root)
+
+# }}}
+
+
+# {{{ create a reshaped view of a numpy array
+
+def reshaped_view(a, newshape):
+    """ Create a new view object with shape ``newshape`` without copying the data of
+    ``a``. This function is different from ``numpy.reshape`` by raising an
+    exception when data copy is necessary.
+
+    :arg a: a :class:`numpy.ndarray` object.
+    :arg newshape: an ``int`` object or a tuple of ``int`` objects.
+
+    .. versionadded:: 2018.4
+    """
+
+    newview = a.view()
+    newview.shape = newshape
+    return newview
+
+# }}}
+
+
+# {{{ process timer
+
+class ProcessTimer(object):
+    """Measures elapsed wall time and process time.
+
+    .. automethod:: __enter__
+    .. automethod:: __exit__
+    .. automethod:: done
+
+    Timing data attributes:
+
+    .. attribute:: wall_elapsed
+    .. attribute:: process_elapsed
+
+        Only available in Python 3.3+.
+
+    .. versionadded:: 2018.5
+    """
+
+    def __init__(self):
+        import time
+        if sys.version_info >= (3, 3):
+            self.perf_counter_start = time.perf_counter()
+            self.process_time_start = time.process_time()
+
+        else:
+            import timeit
+            self.time_start = timeit.default_timer()
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.done()
+
+    def done(self):
+        # pylint: disable=attribute-defined-outside-init
+
+        import time
+        if sys.version_info >= (3, 3):
+            self.wall_elapsed = time.perf_counter() - self.perf_counter_start
+            self.process_elapsed = time.process_time() - self.process_time_start
+
+        else:
+            import timeit
+            self.wall_elapsed = timeit.default_timer() - self.time_start
+            self.process_elapsed = None
+
+# }}}
+
+
+# {{{ log utilities
+
+class ProcessLogger(object):  # pylint: disable=too-many-instance-attributes
+    """Logs the completion time of a (presumably) lengthy process to :mod:`logging`.
+    Only uses a high log level if the process took perceptible time.
+
+    .. automethod:: __init__
+    .. automethod:: done
+    .. automethod:: __enter__
+    .. automethod:: __exit__
+    """
+
+    default_noisy_level = logging.INFO
+
+    def __init__(  # pylint: disable=too-many-arguments
+            self, logger, description,
+            silent_level=None, noisy_level=None, long_threshold_seconds=None):
+        self.logger = logger
+        self.description = description
+        self.silent_level = silent_level or logging.DEBUG
+        self.noisy_level = noisy_level or self.default_noisy_level
+        self.long_threshold_seconds = (
+                # 0 is a valid value that should override the default
+                0.3 if long_threshold_seconds is None else long_threshold_seconds)
+
+        self.logger.log(self.silent_level, "%s: start", self.description)
+        self.is_done = False
+
+        import threading
+        self.late_start_log_thread = threading.Thread(target=self._log_start_if_long)
+        # Do not delay interpreter exit if thread not finished.
+        self.late_start_log_thread.daemon = True
+        self.late_start_log_thread.start()
+
+        self.timer = ProcessTimer()
+
+    def _log_start_if_long(self):
+        from time import sleep
+
+        sleep_duration = 10*self.long_threshold_seconds
+        sleep(sleep_duration)
+
+        if not self.is_done:
+            self.logger.log(
+                    self.noisy_level, "%s: started %.gs ago",
+                    self.description,
+                    sleep_duration)
+
+    def done(  # pylint: disable=keyword-arg-before-vararg
+            self, extra_msg=None, *extra_fmt_args):
+        self.timer.done()
+        self.is_done = True
+
+        wall_elapsed = self.timer.wall_elapsed
+        process_elapsed = self.timer.process_elapsed
+
+        completion_level = (
+                self.noisy_level
+                if wall_elapsed > self.long_threshold_seconds
+                else self.silent_level)
+
+        if process_elapsed is not None:
+            msg = "%s: completed (%.2fs wall, %.1fx CPU)"
+            fmt_args = [self.description, wall_elapsed, process_elapsed/wall_elapsed]
+        else:
+            msg = "%s: completed (%f.2s wall)"
+            fmt_args = [self.description, wall_elapsed]
+
+        if extra_msg:
+            msg += ": " + extra_msg
+            fmt_args.extend(extra_fmt_args)
+
+        self.logger.log(completion_level, msg, *fmt_args)
+
+    def __enter__(self):
+        pass
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.done()
+
+
+class DebugProcessLogger(ProcessLogger):
+    default_noisy_level = logging.DEBUG
+
+
+class log_process(object):  # noqa: N801
+    """A decorator that uses :class:`ProcessLogger` to log data about calls
+    to the wrapped function.
+    """
+
+    def __init__(self, logger, description=None):
+        self.logger = logger
+        self.description = description
+
+    def __call__(self, wrapped):
+        def wrapper(*args, **kwargs):
+            with ProcessLogger(
+                    self.logger,
+                    self.description or wrapped.__name__):
+                return wrapped(*args, **kwargs)
+
+        from functools import update_wrapper
+        new_wrapper = update_wrapper(wrapper, wrapped)
+
+        return new_wrapper
+
+# }}}
+
+
 def _test():
     import doctest
     doctest.testmod()
